@@ -29,7 +29,6 @@ import sys
 
 from nuitka.build.DataComposerInterface import runDataComposer
 from nuitka.constants.Serialization import ConstantAccessor
-from nuitka.finalizations.FinalizeMarkups import getImportedNames
 from nuitka.freezer.IncludedEntryPoints import (
     addIncludedEntryPoints,
     addShlibEntryPoint,
@@ -38,12 +37,20 @@ from nuitka.freezer.IncludedEntryPoints import (
 )
 from nuitka.freezer.Standalone import copyDataFiles
 from nuitka.importing import Importing, Recursion
-from nuitka.Options import getPythonFlags
+from nuitka.Options import hasPythonFlagNoAsserts, hasPythonFlagNoWarnings
 from nuitka.plugins.Plugins import Plugins
 from nuitka.PostProcessing import executePostProcessing
+from nuitka.Progress import (
+    closeProgressBar,
+    reportProgressBar,
+    setupProgressBar,
+)
 from nuitka.PythonVersions import (
     getPythonABI,
     getSupportedPythonVersions,
+    getSystemPrefixPath,
+    isDebianPackagePython,
+    isNuitkaPython,
     python_version,
     python_version_str,
 )
@@ -59,6 +66,8 @@ from nuitka.utils.FileOperations import (
 )
 from nuitka.utils.Importing import getSharedLibrarySuffix
 from nuitka.utils.ModuleNames import ModuleName
+from nuitka.utils.StaticLibraries import getSystemStaticLibPythonPath
+from nuitka.Version import getCommercialVersion, getNuitkaVersion
 
 from . import ModuleRegistry, Options, OutputDirectories, TreeXML
 from .build import SconsInterface
@@ -79,11 +88,12 @@ def _createNodeTree(filename):
 
     """
 
+    # Many cases to deal with, pylint: disable=too-many-branches
+
     # First, build the raw node tree from the source code.
-    main_module = Building.buildModuleTree(
+    main_module = Building.buildMainModuleTree(
         filename=filename,
         package=None,
-        is_top=True,
         is_main=not Options.shallMakeModule(),
     )
 
@@ -101,7 +111,15 @@ def _createNodeTree(filename):
         removeDirectory(path=standalone_dir, ignore_errors=True)
         makePath(standalone_dir)
 
-    deleteFile(path=OutputDirectories.getResultFullpath(), must_exist=False)
+    # Delete result file, to avoid confusion with previous build and to
+    # avoid locking issues after the build.
+    deleteFile(
+        path=OutputDirectories.getResultFullpath(onefile=False), must_exist=False
+    )
+    if Options.isOnefileMode():
+        deleteFile(
+            path=OutputDirectories.getResultFullpath(onefile=True), must_exist=False
+        )
 
     # Second, do it for the directories given.
     for plugin_filename in Options.getShallFollowExtra():
@@ -140,7 +158,8 @@ def _createNodeTree(filename):
 
         if kind != "absolute":
             inclusion_logger.sysexit(
-                "Error, failed to locate module %r you asked to include." % module_name
+                "Error, failed to locate module '%s' you asked to include."
+                % module_name
             )
 
         Recursion.checkPluginSinglePath(
@@ -277,7 +296,7 @@ def makeSourceDirectory():
                 else:
                     ModuleRegistry.removeUncompiledModule(uncompiled_module)
 
-    # Lets check if the recurse-to modules are actually present, and warn the
+    # Lets check if the asked modules are actually present, and warn the
     # user if one of those was not found.
     for any_case_module in Options.getShallFollowModules():
         if "*" in any_case_module or "{" in any_case_module:
@@ -287,31 +306,22 @@ def makeSourceDirectory():
             if module.getFullName() == any_case_module:
                 break
         else:
-            general.warning("Not recursing to unused '%s'." % any_case_module)
+            general.warning(
+                "Did not follow import to unused '%s', consider include options."
+                % any_case_module
+            )
 
     # Prepare code generation, i.e. execute finalization for it.
     for module in ModuleRegistry.getDoneModules():
         if module.isCompiledPythonModule():
             Finalization.prepareCodeGeneration(module)
 
-    # Pick filenames.
-    source_dir = OutputDirectories.getSourceDirectoryPath()
+    # Do some reporting and determine compiled module to work on
+    compiled_modules = []
 
-    module_filenames = pickSourceFilenames(
-        source_dir=source_dir, modules=ModuleRegistry.getDoneModules()
-    )
-
-    # Generate code for modules.
     for module in ModuleRegistry.getDoneModules():
         if module.isCompiledPythonModule():
-            c_filename = module_filenames[module]
-
-            source_code = CodeGeneration.generateModuleCode(
-                module=module,
-                data_filename=os.path.basename(c_filename + "onst"),  # Really .const
-            )
-
-            writeSourceCode(filename=c_filename, source_code=source_code)
+            compiled_modules.append(module)
 
             if Options.isShowInclusion():
                 inclusion_logger.info(
@@ -332,6 +342,37 @@ def makeSourceDirectory():
         else:
             assert False, module
 
+    # Pick filenames.
+    source_dir = OutputDirectories.getSourceDirectoryPath()
+
+    module_filenames = pickSourceFilenames(
+        source_dir=source_dir, modules=compiled_modules
+    )
+
+    setupProgressBar(
+        stage="C Source Generation",
+        unit="module",
+        total=len(compiled_modules),
+    )
+
+    # Generate code for compiled modules, this can be slow, so do it separately
+    # with a progress bar.
+    for module in compiled_modules:
+        c_filename = module_filenames[module]
+
+        reportProgressBar(
+            item=module.getFullName(),
+        )
+
+        source_code = CodeGeneration.generateModuleCode(
+            module=module,
+            data_filename=os.path.basename(c_filename + "onst"),  # Really .const
+        )
+
+        writeSourceCode(filename=c_filename, source_code=source_code)
+
+    closeProgressBar()
+
     (
         helper_decl_code,
         helper_impl_code,
@@ -357,14 +398,26 @@ def makeSourceDirectory():
         source_code=constants_body_code,
     )
 
-    for filename, source_code in Plugins.getExtraCodeFiles().items():
-        target_dir = os.path.join(source_dir, "plugins")
 
-        if not os.path.isdir(target_dir):
-            makePath(target_dir)
+def runPgoBinary():
+    # Exit code, we do not insist on anything.
+    Execution.call(
+        [
+            Options.getPgoExecutable()
+            or os.path.abspath(OutputDirectories.getResultFullpath(onefile=False))
+        ]
+        + Options.getPgoArgs(),
+        shell=False,
+    )
 
-        writeSourceCode(
-            filename=os.path.join(target_dir, filename), source_code=source_code
+    # TODO: For other compilers, this will have to be different suffix, maybe check all of them
+    constants_pgo_filename = os.path.join(
+        OutputDirectories.getSourceDirectoryPath(), "__constants.gcda"
+    )
+
+    if not os.path.exists(constants_pgo_filename):
+        general.sysexit(
+            "Error, no PGO information produced, did the created binary run at all?"
         )
 
 
@@ -378,6 +431,7 @@ def runSconsBackend(quiet):
     options = {
         "result_name": OutputDirectories.getResultBasepath(onefile=False),
         "source_dir": OutputDirectories.getSourceDirectoryPath(),
+        "nuitka_python": asBoolStr(isNuitkaPython()),
         "debug_mode": asBoolStr(Options.is_debug),
         "python_debug": asBoolStr(Options.isPythonDebug()),
         "unstripped_mode": asBoolStr(Options.isUnstripped()),
@@ -387,7 +441,7 @@ def runSconsBackend(quiet):
         "trace_mode": asBoolStr(Options.shallTraceExecution()),
         "python_version": python_version_str,
         "target_arch": Utils.getArchitecture(),
-        "python_prefix": getDirectoryRealPath(sys.prefix),
+        "python_prefix": getDirectoryRealPath(getSystemPrefixPath()),
         "nuitka_src": SconsInterface.getSconsDataPath(),
         "module_count": "%d"
         % (
@@ -398,25 +452,29 @@ def runSconsBackend(quiet):
     }
 
     if not Options.shallMakeModule():
-        options["result_exe"] = OutputDirectories.getResultFullpath()
-
-    # Ask Scons to cache on Windows, except where the directory is thrown
-    # away. On non-Windows you can should use ccache instead.
-    if not Options.isRemoveBuildDir() and Utils.getOS() == "Windows":
-        options["cache_mode"] = "true"
-
-    if Options.isLto():
-        options["lto_mode"] = asBoolStr(True)
+        options["result_exe"] = OutputDirectories.getResultFullpath(onefile=False)
 
     if Options.shallUseStaticLibPython():
-        options["static_libpython"] = asBoolStr(True)
+        options["static_libpython"] = getSystemStaticLibPythonPath()
 
-    if Options.shallDisableConsoleWindow():
-        options["win_disable_console"] = asBoolStr(True)
+    if isDebianPackagePython():
+        options["debian_python"] = asBoolStr(True)
 
     if Options.isStandaloneMode():
         options["standalone_mode"] = asBoolStr(True)
 
+    if Options.isOnefileMode():
+        options["onefile_mode"] = asBoolStr(True)
+
+        if Options.isOnefileTempDirMode():
+            options["onefile_temp_mode"] = asBoolStr(True)
+
+    if Options.getForcedStdoutPath():
+        options["forced_stdout_path"] = Options.getForcedStdoutPath()
+
+    if Options.getForcedStderrPath():
+        options["forced_stderr_path"] = Options.getForcedStderrPath()
+
     if Options.shallTreatUninstalledPython():
         options["uninstalled_python"] = asBoolStr(True)
 
@@ -425,34 +483,16 @@ def runSconsBackend(quiet):
             len(ModuleRegistry.getUncompiledTechnicalModules())
         )
 
-    if Options.isShowScons():
-        options["show_scons"] = asBoolStr(True)
-
-    if Options.isMingw64():
-        options["mingw_mode"] = asBoolStr(True)
-
-    if Options.getMsvcVersion():
-        msvc_version = Options.getMsvcVersion()
-
-        msvc_version = msvc_version.replace("exp", "Exp")
-        if "." not in msvc_version:
-            msvc_version += ".0"
-
-        options["msvc_version"] = msvc_version
-
     if Utils.getOS() == "Windows":
         options["noelf_mode"] = asBoolStr(True)
 
-    if Options.isClang():
-        options["clang_mode"] = asBoolStr(True)
-
     if Options.isProfile():
         options["profile_mode"] = asBoolStr(True)
 
-    if "no_warnings" in getPythonFlags():
+    if hasPythonFlagNoWarnings():
         options["no_python_warnings"] = asBoolStr(True)
 
-    if "no_asserts" in getPythonFlags():
+    if hasPythonFlagNoAsserts():
         options["python_sysflag_optimize"] = asBoolStr(True)
 
     if python_version < 0x300 and sys.flags.py3k_warning:
@@ -466,7 +506,7 @@ def runSconsBackend(quiet):
     if sys.flags.bytes_warning:
         options["python_sysflag_bytes_warning"] = asBoolStr(True)
 
-    if int(os.environ.get("NUITKA_SITE_FLAG", "no_site" in Options.getPythonFlags())):
+    if int(os.environ.get("NUITKA_SITE_FLAG", Options.hasPythonFlagNoSite())):
         options["python_sysflag_no_site"] = asBoolStr(True)
 
     if "trace_imports" in Options.getPythonFlags():
@@ -485,25 +525,21 @@ def runSconsBackend(quiet):
     if abiflags:
         options["abiflags"] = abiflags
 
-    cpp_defines = Plugins.getPreprocessorSymbols()
-    if cpp_defines:
-        options["cpp_defines"] = ",".join(
-            "%s%s%s" % (key, "=" if value else "", value or "")
-            for key, value in cpp_defines.items()
-        )
-
-    link_libraries = Plugins.getExtraLinkLibraries()
-    if link_libraries:
-        options["link_libraries"] = ",".join(link_libraries)
-
     if Options.shallMakeModule():
         options["module_suffix"] = getSharedLibrarySuffix(preferred=True)
 
-    if Options.shallRunInDebugger():
-        options["full_names"] = asBoolStr(True)
+    SconsInterface.setCommonOptions(options)
 
-    if Options.assumeYesForDownloads():
-        options["assume_yes_for_downloads"] = asBoolStr(True)
+    # For PGO, we have a 2 pass system. TODO: Make it more global for onefile
+    # and standalone mode proper support, which might need data files to be
+    # there, which currently are not yet there, so it won't run.
+    if Options.isPgoMode():
+        options["pgo_mode"] = "generate"
+        SconsInterface.runScons(
+            options=options, quiet=quiet, scons_filename="Backend.scons"
+        )
+        runPgoBinary()
+        options["pgo_mode"] = "use"
 
     return (
         SconsInterface.runScons(
@@ -629,9 +665,18 @@ def compileTree():
 
     general.info("Running data composer tool for optimal constant value handling.")
 
-    # TODO: On Windows, we could run this in parallel to Scons, on Linux we need it
-    # for linking.
-    runDataComposer(source_dir)
+    blob_filename = runDataComposer(source_dir)
+    Plugins.onDataComposerResult(blob_filename)
+
+    for filename, source_code in Plugins.getExtraCodeFiles().items():
+        target_dir = os.path.join(source_dir, "plugins")
+
+        if not os.path.isdir(target_dir):
+            makePath(target_dir)
+
+        writeSourceCode(
+            filename=os.path.join(target_dir, filename), source_code=source_code
+        )
 
     general.info("Running C level backend compilation via Scons.")
 
@@ -681,7 +726,10 @@ def main():
     # Main has to fulfill many options, leading to many branches and statements
     # to deal with them.  pylint: disable=too-many-branches,too-many-statements
     if not Options.shallDumpBuiltTreeXML():
-        general.info("Starting Python compilation.")
+        general.info(
+            "Starting Python compilation with Nuitka %r on Python %r commercial %r."
+            % (getNuitkaVersion(), python_version_str, getCommercialVersion())
+        )
 
     filename = Options.getPositionalArgs()[0]
 
@@ -717,38 +765,6 @@ def main():
 
         executePostProcessing()
 
-        if Options.shallMakeModule() and Options.shallCreatePyiFile():
-            pyi_filename = OutputDirectories.getResultBasepath() + ".pyi"
-
-            putTextFileContents(
-                filename=pyi_filename,
-                contents="""\
-# This file was generated by Nuitka and describes the types of the
-# created shared library.
-
-# At this time it lists only the imports made and can be used by the
-# tools that bundle libraries, including Nuitka itself. For instance
-# standalone mode usage of the created library will need it.
-
-# In the future, this will also contain type information for values
-# in the module, so IDEs will use this. Therefore please include it
-# when you make software releases of the extension module that it
-# describes.
-
-%(imports)s
-
-# This is not Python source even if it looks so. Make it clear for
-# now. This was decided by PEP 484 designers.
-__name__ = ...
-
-"""
-                % {
-                    "imports": "\n".join(
-                        "import %s" % module_name for module_name in getImportedNames()
-                    )
-                },
-            )
-
         if Options.isStandaloneMode():
             binary_filename = options["result_exe"]
 
@@ -803,6 +819,12 @@ __name__ = ...
 
         # Execute the module immediately if option was given.
         if Options.shallExecuteImmediately():
+            run_filename = OutputDirectories.getResultRunFilename(
+                onefile=Options.isOnefileMode()
+            )
+
+            general.info("Launching %r." % run_filename)
+
             if Options.shallMakeModule():
                 executeModule(
                     tree=main_module,
@@ -810,8 +832,6 @@ __name__ = ...
                 )
             else:
                 executeMain(
-                    binary_filename=OutputDirectories.getResultFullpath(
-                        onefile=Options.isOnefileMode()
-                    ),
+                    binary_filename=run_filename,
                     clean_path=Options.shallClearPythonPathEnvironment(),
                 )
