@@ -38,8 +38,14 @@ from nuitka.utils.ThreadedExecutor import RLock
 # Written by Options module.
 is_quiet = False
 
-# Written by Options module
-use_progressbar = False
+# We have to interact with displayed progress bars when doing out trace outputs.
+progress = None
+
+
+def setQuiet():
+    # singleton, pylint: disable=global-statement
+    global is_quiet
+    is_quiet = True
 
 
 def printIndented(level, *what):
@@ -111,7 +117,7 @@ trace_lock = RLock()
 
 @contextmanager
 def withTraceLock():
-    """ Hold a lock, so traces cannot be output at the same time mixing them up. """
+    """Hold a lock, so traces cannot be output at the same time mixing them up."""
 
     trace_lock.acquire()
     yield
@@ -126,9 +132,11 @@ def my_print(*args, **kwargs):
     Use kwarg style=[option] to print in a style listed below
     """
 
-    global progress  # singleton, pylint: disable=global-statement
-    if progress is not None:
-        progress.clear()
+    file_output = kwargs.get("file", sys.stdout)
+    is_atty = file_output.isatty()
+
+    if progress and is_atty:
+        progress.hideProgressBar()
 
     with withTraceLock():
         if "style" in kwargs:
@@ -141,7 +149,7 @@ def my_print(*args, **kwargs):
             else:
                 end = "\n"
 
-            if style is not None and sys.stdout.isatty():
+            if style is not None and is_atty:
                 enable_style = getEnableStyleCode(style)
 
                 if enable_style is None:
@@ -155,32 +163,46 @@ def my_print(*args, **kwargs):
 
             print(*args, end=end, **kwargs)
 
-            if style is not None and sys.stdout.isatty():
+            if style is not None and is_atty:
                 print(getDisableStyleCode(), end="", **kwargs)
         else:
             print(*args, **kwargs)
 
         # Flush the output.
-        kwargs.get("file", sys.stdout).flush()
+        file_output.flush()
+
+    if progress and is_atty:
+        progress.resumeProgressBar()
 
 
 class OurLogger(object):
-    def __init__(self, name, base_style=None):
+    def __init__(self, name, quiet=False, base_style=None):
         self.name = name
         self.base_style = base_style
-        self.is_quiet = False
+        self.is_quiet = quiet
+
+        # Can disable warnings, we do that for options parsing during re-execution.
+        self.is_no_warnings = False
 
     def my_print(self, message, **kwargs):
         # For overload, pylint: disable=no-self-use
         my_print(message, **kwargs)
 
     def warning(self, message, style="red"):
-        message = "%s:WARNING: %s" % (self.name, message)
+        if not self.is_no_warnings:
+            if self.name:
+                message = "%s:WARNING: %s" % (self.name, message)
+            else:
+                message = "WARNING: %s" % message
 
-        style = style or self.base_style
-        self.my_print(message, style=style, file=sys.stderr)
+            style = style or self.base_style
+            self.my_print(message, style=style, file=sys.stderr)
 
     def sysexit(self, message, exit_code=1):
+        from nuitka.Progress import closeProgressBar
+
+        closeProgressBar()
+
         self.my_print("FATAL: %s" % message, style="red", file=sys.stderr)
 
         sys.exit(exit_code)
@@ -204,40 +226,46 @@ class OurLogger(object):
 
 
 class FileLogger(OurLogger):
-    def __init__(self, name, base_style=None, file_handle=sys.stdout):
-        OurLogger.__init__(self, name=name, base_style=base_style)
+    def __init__(self, name, quiet=False, base_style=None, file_handle=None):
+        OurLogger.__init__(self, name=name, quiet=quiet, base_style=base_style)
 
         self.file_handle = file_handle
 
     def my_print(self, message, **kwargs):
-        message = message + "\n"
+        if "file" not in kwargs:
+            kwargs["file"] = self.file_handle or sys.stdout
 
-        self.file_handle.write(message)
-        self.file_handle.flush()
+        my_print(message, **kwargs)
+        kwargs["file"].flush()
 
     def setFileHandle(self, file_handle):
         self.file_handle = file_handle
 
     def info(self, message, style=None):
-        if not self.isQuiet() or self.file_handle is not sys.stdout:
+        if not self.isQuiet() or self.file_handle:
             message = "%s:INFO: %s" % (self.name, message)
 
             style = style or self.base_style
             self.my_print(message, style=style)
 
     def debug(self, message, style=None):
-        if self.file_handle is not sys.stdout:
+        if self.file_handle:
             message = "%s:DEBUG: %s" % (self.name, message)
 
             style = style or self.base_style
             self.my_print(message, style=style)
 
+    def info_fileoutput(self, message, other_logger, style=None):
+        if self.file_handle:
+            self.info(message, style=style)
+        else:
+            other_logger.info(message, style=style)
+
 
 general = OurLogger("Nuitka")
-codegen_missing = OurLogger("Nuitka-codegen-missing")
 plugins_logger = OurLogger("Nuitka-Plugins")
 recursion_logger = OurLogger("Nuitka-Recursion")
-progress_logger = OurLogger("Nuitka-Progress")
+progress_logger = OurLogger("Nuitka-Progress", quiet=True)
 memory_logger = OurLogger("Nuitka-Memory")
 dependencies_logger = OurLogger("Nuitka-Dependencies")
 optimization_logger = FileLogger("Nuitka-Optimization")
@@ -249,24 +277,4 @@ postprocessing_logger = OurLogger("Nuitka-Postprocessing")
 options_logger = OurLogger("Nuitka-Options")
 unusual_logger = OurLogger("Nuitka-Unusual")
 datacomposer_logger = OurLogger("Nuitka-Datacomposer")
-
-progress = None
-
-
-def reportProgressBar(stage, unit, item, total):
-    global progress  # singleton, pylint: disable=global-statement
-
-    if progress is None and use_progressbar:
-        try:
-            from tqdm import tqdm
-
-            progress = tqdm(initial=1, total=total, unit=unit)
-        except ImportError:
-            return
-
-    if progress is not None:
-        progress.set_description(stage)
-        progress.set_postfix(item=item)
-        progress.unit = unit
-        progress.total = total
-        progress.update(1)
+onefile_logger = OurLogger("Nuitka-Onefile")
