@@ -1,4 +1,4 @@
-#     Copyright 2018, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -30,15 +30,19 @@ from nuitka.Builtins import (
     builtin_exception_names,
     builtin_exception_values,
     builtin_names,
-    builtin_type_names
+    builtin_type_names,
 )
-from nuitka.Options import getPythonFlags
+from nuitka.Options import hasPythonFlagNoAsserts
 from nuitka.PythonVersions import python_version
 from nuitka.specs import BuiltinParameterSpecs
 
 from .ConstantRefNodes import makeConstantRefNode
-from .ExceptionNodes import ExpressionBuiltinMakeException
+from .ExceptionNodes import (
+    ExpressionBuiltinMakeException,
+    ExpressionBuiltinMakeExceptionImportError,
+)
 from .ExpressionBases import CompileTimeConstantExpressionBase
+from .shapes.BuiltinTypeShapes import tshape_exception_class
 
 
 class ExpressionBuiltinRefBase(CompileTimeConstantExpressionBase):
@@ -47,10 +51,7 @@ class ExpressionBuiltinRefBase(CompileTimeConstantExpressionBase):
     __slots__ = ("builtin_name",)
 
     def __init__(self, builtin_name, source_ref):
-        CompileTimeConstantExpressionBase.__init__(
-            self,
-            source_ref = source_ref
-        )
+        CompileTimeConstantExpressionBase.__init__(self, source_ref=source_ref)
 
         self.builtin_name = builtin_name
 
@@ -58,57 +59,61 @@ class ExpressionBuiltinRefBase(CompileTimeConstantExpressionBase):
         del self.parent
 
     def getDetails(self):
-        return {
-            "builtin_name" : self.builtin_name
-        }
+        return {"builtin_name": self.builtin_name}
 
     def getBuiltinName(self):
         return self.builtin_name
 
-    def isKnownToBeHashable(self):
+    @staticmethod
+    def isKnownToBeHashable():
         return True
 
-    def mayRaiseException(self, exception_type):
+    @staticmethod
+    def mayRaiseException(exception_type):
         return False
 
-    def mayHaveSideEffects(self):
+    @staticmethod
+    def mayHaveSideEffects():
         return False
 
     def getStrValue(self):
         return makeConstantRefNode(
-            constant      = str(self.getCompileTimeConstant()),
-            user_provided = True,
-            source_ref    = self.getSourceReference(),
+            constant=str(self.getCompileTimeConstant()),
+            user_provided=True,
+            source_ref=self.source_ref,
         )
 
-_debug_value = "no_asserts" not in getPythonFlags()
 
-def makeExpressionBuiltinRef(builtin_name, source_ref):
+def makeExpressionBuiltinTypeRef(builtin_name, source_ref):
+    return makeConstantRefNode(
+        constant=__builtins__[builtin_name], source_ref=source_ref
+    )
+
+
+quick_names = {"None": None, "True": True, "False": False, "Ellipsis": Ellipsis}
+
+
+def makeExpressionBuiltinRef(builtin_name, locals_scope, source_ref):
     assert builtin_name in builtin_names, builtin_name
 
-    quick_names = {
-        "None"      : None,
-        "True"      : True,
-        "False"     : False,
-        "__debug__" : _debug_value,
-        "Ellipsis"  : Ellipsis,
-    }
-
     if builtin_name in quick_names:
         return makeConstantRefNode(
-            constant   = quick_names[builtin_name],
-            source_ref = source_ref
+            constant=quick_names[builtin_name], source_ref=source_ref
+        )
+    elif builtin_name == "__debug__":
+        return makeConstantRefNode(
+            constant=not hasPythonFlagNoAsserts(), source_ref=source_ref
         )
     elif builtin_name in builtin_type_names:
-        return makeConstantRefNode(
-            constant   = __builtins__[builtin_name],
-            source_ref = source_ref
+        return makeExpressionBuiltinTypeRef(
+            builtin_name=builtin_name, source_ref=source_ref
+        )
+    elif builtin_name in ("dir", "eval", "exec", "execfile", "locals", "vars"):
+        return ExpressionBuiltinWithContextRef(
+            builtin_name=builtin_name, locals_scope=locals_scope, source_ref=source_ref
         )
     else:
-        return ExpressionBuiltinRef(
-            builtin_name = builtin_name,
-            source_ref   = source_ref
-        )
+        return ExpressionBuiltinRef(builtin_name=builtin_name, source_ref=source_ref)
 
 
 class ExpressionBuiltinRef(ExpressionBuiltinRefBase):
@@ -116,49 +121,70 @@ class ExpressionBuiltinRef(ExpressionBuiltinRefBase):
 
     __slots__ = ()
 
+    # For overload
+    locals_scope = None
+
+    @staticmethod
+    def isExpressionBuiltinRef():
+        return True
+
     def __init__(self, builtin_name, source_ref):
         ExpressionBuiltinRefBase.__init__(
-            self,
-            builtin_name = builtin_name,
-            source_ref   = source_ref
+            self, builtin_name=builtin_name, source_ref=source_ref
         )
 
-    def isCompileTimeConstant(self):
-        return True
-
     def getCompileTimeConstant(self):
         return __builtins__[self.builtin_name]
 
     def computeExpressionRaw(self, trace_collection):
         return self, None, None
 
-    def computeExpressionCall(self, call_node, call_args, call_kw,
-                              trace_collection):
-        from nuitka.optimizations.OptimizeBuiltinCalls import computeBuiltinCall
+    def computeExpressionCall(self, call_node, call_args, call_kw, trace_collection):
+        from nuitka.optimizations.OptimizeBuiltinCalls import (
+            computeBuiltinCall,
+        )
 
         # Anything may happen. On next pass, if replaced, we might be better
         # but not now.
         trace_collection.onExceptionRaiseExit(BaseException)
 
         new_node, tags, message = computeBuiltinCall(
-            builtin_name = self.builtin_name,
-            call_node    = call_node
+            builtin_name=self.builtin_name, call_node=call_node
         )
 
         if self.builtin_name in ("dir", "eval", "exec", "execfile", "locals", "vars"):
             # Just inform the collection that all has escaped.
-            trace_collection.onLocalsUsage(self.getParentVariableProvider())
+            trace_collection.onLocalsUsage(locals_scope=self.getLocalsScope())
 
         return new_node, tags, message
 
-    def getStringValue(self):
-        return repr(self.getCompileTimeConstant())
-
-    def isKnownToBeIterable(self, count):
+    @staticmethod
+    def isKnownToBeIterable(count):
         # TODO: Why yes, some may be, could be told here.
         return None
 
 
+class ExpressionBuiltinWithContextRef(ExpressionBuiltinRef):
+    """Same as ExpressionBuiltinRef, but with a context it refers to."""
+
+    kind = "EXPRESSION_BUILTIN_WITH_CONTEXT_REF"
+
+    __slots__ = ("locals_scope",)
+
+    def __init__(self, builtin_name, locals_scope, source_ref):
+        ExpressionBuiltinRef.__init__(
+            self, builtin_name=builtin_name, source_ref=source_ref
+        )
+
+        self.locals_scope = locals_scope
+
+    def getDetails(self):
+        return {"builtin_name": self.builtin_name, "locals_scope": self.locals_scope}
+
+    def getLocalsScope(self):
+        return self.locals_scope
+
+
 class ExpressionBuiltinAnonymousRef(ExpressionBuiltinRefBase):
     kind = "EXPRESSION_BUILTIN_ANONYMOUS_REF"
 
@@ -168,23 +194,15 @@ class ExpressionBuiltinAnonymousRef(ExpressionBuiltinRefBase):
         assert builtin_name in builtin_anon_names, builtin_name
 
         ExpressionBuiltinRefBase.__init__(
-            self,
-            builtin_name = builtin_name,
-            source_ref   = source_ref
+            self, builtin_name=builtin_name, source_ref=source_ref
         )
 
-    def isCompileTimeConstant(self):
-        return True
-
     def getCompileTimeConstant(self):
         return builtin_anon_names[self.builtin_name]
 
     def computeExpressionRaw(self, trace_collection):
         return self, None, None
 
-    def getStringValue(self):
-        return repr(self.getCompileTimeConstant())
-
 
 class ExpressionBuiltinExceptionRef(ExpressionBuiltinRefBase):
     kind = "EXPRESSION_BUILTIN_EXCEPTION_REF"
@@ -195,22 +213,20 @@ class ExpressionBuiltinExceptionRef(ExpressionBuiltinRefBase):
         assert exception_name in builtin_exception_names, exception_name
 
         ExpressionBuiltinRefBase.__init__(
-            self,
-            builtin_name = exception_name,
-            source_ref   = source_ref
+            self, builtin_name=exception_name, source_ref=source_ref
         )
 
     def getDetails(self):
-        return {
-            "exception_name" : self.builtin_name
-        }
+        return {"exception_name": self.builtin_name}
 
     getExceptionName = ExpressionBuiltinRefBase.getBuiltinName
 
-    def isCompileTimeConstant(self):
-        return True
+    @staticmethod
+    def getTypeShape():
+        return tshape_exception_class
 
-    def mayRaiseException(self, exception_type):
+    @staticmethod
+    def mayRaiseException(exception_type):
         return False
 
     def getCompileTimeConstant(self):
@@ -220,34 +236,35 @@ class ExpressionBuiltinExceptionRef(ExpressionBuiltinRefBase):
         # Not much that can be done here.
         return self, None, None
 
-    def computeExpressionCall(self, call_node, call_args, call_kw,
-                              trace_collection):
+    def computeExpressionCall(self, call_node, call_args, call_kw, trace_collection):
         exception_name = self.getExceptionName()
 
-        # TODO: Keyword only arguments of it, are not properly handled yet by
-        # the built-in call code.
-        if exception_name == "ImportError" and python_version >= 300:
-            if call_kw is not None and \
-               (not call_kw.isExpressionConstantRef() or call_kw.getConstant() != {}):
-                return call_node, None, None
+        def createBuiltinMakeException(args, name=None, path=None, source_ref=None):
+            if exception_name == "ImportError" and python_version >= 0x300:
+                return ExpressionBuiltinMakeExceptionImportError(
+                    exception_name=exception_name,
+                    args=args,
+                    name=name,
+                    path=path,
+                    source_ref=source_ref,
+                )
+            else:
+                # We expect to only get the star arguments for these.
+                assert name is None
+                assert path is None
 
-        def createBuiltinMakeException(args, source_ref):
-            return ExpressionBuiltinMakeException(
-                exception_name = exception_name,
-                args           = args,
-                source_ref     = source_ref
-            )
+                return ExpressionBuiltinMakeException(
+                    exception_name=exception_name, args=args, source_ref=source_ref
+                )
 
         new_node = BuiltinParameterSpecs.extractBuiltinArgs(
-            node          = call_node,
-            builtin_class = createBuiltinMakeException,
-            builtin_spec  = BuiltinParameterSpecs.makeBuiltinExceptionParameterSpec(
-                exception_name = exception_name
-            )
+            node=call_node,
+            builtin_class=createBuiltinMakeException,
+            builtin_spec=BuiltinParameterSpecs.makeBuiltinExceptionParameterSpec(
+                exception_name=exception_name
+            ),
         )
 
-        # TODO: Don't allow this to happen.
-        if new_node is None:
-            return call_node, None, None
+        assert new_node is not None
 
         return new_node, "new_expression", "Detected built-in exception making."
