@@ -27,29 +27,34 @@ The base class in PluginBase will serve as documentation of available.
 
 """
 
+import inspect
 import os
 import pkgutil
 import shutil
-from optparse import OptionGroup
+from optparse import OptionConflictError, OptionGroup
 
 import nuitka.plugins.commercial
 import nuitka.plugins.standard
-from nuitka import Options
+from nuitka import Options, OutputDirectories
 from nuitka.__past__ import basestring  # pylint: disable=I0021,redefined-builtin
+from nuitka.build.DataComposerInterface import deriveModuleConstantsBlobName
 from nuitka.containers.odict import OrderedDict
 from nuitka.containers.oset import OrderedSet
+from nuitka.Errors import NuitkaPluginError
 from nuitka.freezer.IncludedEntryPoints import makeDllEntryPointOld
 from nuitka.ModuleRegistry import addUsedModule
 from nuitka.Tracing import plugins_logger, printLine
-from nuitka.utils.FileOperations import makePath
+from nuitka.utils.FileOperations import makePath, putTextFileContents, relpath
 from nuitka.utils.Importing import importFileAsModule
 from nuitka.utils.ModuleNames import ModuleName
 
 from .PluginBase import NuitkaPluginBase, post_modules, pre_modules
 
+# Maps plugin name to plugin instances.
 active_plugins = OrderedDict()
 plugin_name2plugin_classes = {}
 plugin_options = {}
+plugin_datatag2pluginclasses = {}
 plugin_values = {}
 user_plugins = OrderedSet()
 
@@ -86,6 +91,17 @@ def getActivePlugins():
     return active_plugins.values()
 
 
+def getActiveQtPlugin():
+    from .standard.PySidePyQtPlugin import getQtPluginNames
+
+    for plugin_name in getQtPluginNames():
+        if hasActivePlugin(plugin_name):
+            if hasActivePlugin(plugin_name):
+                return plugin_name
+
+    return None
+
+
 def hasActivePlugin(plugin_name):
     """Decide if a plugin is active.
 
@@ -111,6 +127,9 @@ def getPluginClass(plugin_name):
     # First, load plugin classes, to know what we are talking about.
     loadPlugins()
 
+    # Backward compatibility.
+    plugin_name = Options.getPluginNameConsideringRenames(plugin_name)
+
     if plugin_name not in plugin_name2plugin_classes:
         plugins_logger.sysexit("Error, unknown plug-in '%s' referenced." % plugin_name)
 
@@ -162,7 +181,11 @@ def _loadPluginClassesFromPath(scan_path):
             assert detector.plugin_name is None, detector
             detector.plugin_name = plugin_class.plugin_name
 
-            assert plugin_class in plugin_classes
+            if plugin_class not in plugin_classes:
+                plugins_logger.sysexit(
+                    "Plugin detector %r references unknown plugin %r"
+                    % (detector, plugin_class)
+                )
 
             plugin_classes.remove(detector)
             plugin_classes.remove(plugin_class)
@@ -181,7 +204,7 @@ def loadStandardPluginClasses():
 
     Notes:
         Scan through the 'standard' and 'commercial' sub-folder of the folder
-        where this script resides. Import each valid Python module (but not
+        where this module resides. Import each valid Python module (but not
         packages) and process it as a plugin.
     Returns:
         None
@@ -191,18 +214,106 @@ def loadStandardPluginClasses():
 
 
 class Plugins(object):
-    @staticmethod
-    def isPluginActive(plugin_name):
-        """ Is a plugin activated. """
-
-        return plugin_name in active_plugins
+    implicit_imports_cache = {}
 
     @staticmethod
-    def considerImplicitImports(module, signal_change):
+    def _considerImplicitImports(plugin, module):
+        from nuitka.importing import Importing
+
+        result = []
+
+        for full_name in plugin.getImplicitImports(module):
+            if type(full_name) in (tuple, list):
+                raise NuitkaPluginError(
+                    "Plugin %r needs to be change to only return modules names, not %r"
+                    % (plugin, full_name)
+                )
+
+            full_name = ModuleName(full_name)
+
+            try:
+                _module_package, module_filename, _finding = Importing.findModule(
+                    importing=module,
+                    module_name=full_name,
+                    parent_package=None,
+                    level=-1,
+                    warn=False,
+                )
+
+                module_filename = plugin.locateModule(
+                    importing=module, module_name=full_name
+                )
+            except Exception:
+                plugin.warning(
+                    "Problem locating '%s' for implicit imports of '%s'."
+                    % (module.getFullName(), full_name)
+                )
+                raise
+
+            if module_filename is None:
+                if Options.isShowInclusion():
+                    plugin.info(
+                        "Implicit module '%s' suggested for '%s' not found."
+                        % (full_name, module.getFullName())
+                    )
+
+                continue
+
+            result.append((full_name, module_filename))
+
+        if result:
+            plugin.info(
+                "Implicit dependencies of module '%s' added '%s'."
+                % (module.getFullName(), ",".join(r[0] for r in result))
+            )
+
+        return result
+
+    @staticmethod
+    def _reportImplicitImports(implicit_imports, signal_change):
+        from nuitka.importing import Recursion
+        from nuitka.importing.Importing import getModuleNameAndKindFromFilename
+
+        for full_name, module_filename in implicit_imports:
+            _module_name2, module_kind = getModuleNameAndKindFromFilename(
+                module_filename
+            )
+
+            # This will get back to all other plugins allowing them to inhibit it though.
+            decision, reason = Recursion.decideRecursion(
+                module_filename=module_filename,
+                module_name=full_name,
+                module_kind=module_kind,
+            )
+
+            if decision:
+                imported_module = Recursion.recurseTo(
+                    signal_change=signal_change,
+                    module_package=full_name.getPackageName(),
+                    module_filename=module_filename,
+                    module_relpath=relpath(module_filename),
+                    module_kind=module_kind,
+                    reason=reason,
+                )
+
+                addUsedModule(imported_module)
+
+    @classmethod
+    def considerImplicitImports(cls, module, signal_change):
         for plugin in getActivePlugins():
-            plugin.considerImplicitImports(module, signal_change)
+            key = (module.getFullName(), plugin)
 
-        # Post load code may have been created, if so indicate it's used.
+            if key not in cls.implicit_imports_cache:
+                cls.implicit_imports_cache[key] = tuple(
+                    cls._considerImplicitImports(plugin=plugin, module=module)
+                )
+
+            cls._reportImplicitImports(
+                implicit_imports=cls.implicit_imports_cache[key],
+                signal_change=signal_change,
+            )
+
+        # Pre and post load code may have been created, if so indicate it's used.
         full_name = module.getFullName()
 
         if full_name in pre_modules:
@@ -273,6 +384,21 @@ class Plugins(object):
 
         return result
 
+    @staticmethod
+    def getModuleSpecificDllPaths(module_name):
+        """Provide a list of directories, where DLLs should be searched for this package (or module).
+
+        Args:
+            module_name: name of a package or module, for which the DLL path addition applies.
+
+        """
+        result = OrderedSet()
+        for plugin in getActivePlugins():
+            for dll_path in plugin.getModuleSpecificDllPaths(module_name):
+                result.add(dll_path)
+
+        return result
+
     @staticmethod
     def removeDllDependencies(dll_filename, dll_filenames):
         """Create list of removable shared libraries by scanning through the plugins.
@@ -285,15 +411,24 @@ class Plugins(object):
         """
         dll_filenames = tuple(sorted(dll_filenames))
 
-        result = []
+        to_remove = OrderedSet()
 
         for plugin in getActivePlugins():
-            for removed_dll in plugin.removeDllDependencies(
-                dll_filename, dll_filenames
-            ):
-                result.append(removed_dll)
+            removed_dlls = tuple(
+                plugin.removeDllDependencies(dll_filename, dll_filenames)
+            )
 
-        return result
+            if removed_dlls and Options.isShowInclusion():
+                plugin.info(
+                    "Removing DLLs %s of %s by plugin decision."
+                    % (dll_filename, removed_dlls)
+                )
+
+            for removed_dll in removed_dlls:
+                to_remove.add(removed_dll)
+
+        for removed in to_remove:
+            dll_filenames.discard(removed)
 
     @staticmethod
     def considerDataFiles(module):
@@ -310,18 +445,155 @@ class Plugins(object):
                 if value:
                     yield plugin, value
 
-    @staticmethod
-    def onModuleDiscovered(module):
+    @classmethod
+    def _createTriggerLoadedModule(cls, module, trigger_name, code, flags):
+        """Create a "trigger" for a module to be imported.
+
+        Notes:
+            The trigger will incorporate the code to be prepended / appended.
+            Called by @onModuleDiscovered.
+
+        Args:
+            module: the module object (serves as dict key)
+            trigger_name: string ("-preload"/"-postload")
+            code: the code string
+
+        Returns
+            trigger_module
+        """
+
+        from nuitka.tree.Building import buildModule
+
+        module_name = ModuleName(module.getFullName() + trigger_name)
+
+        # In debug mode, put the files in the build folder, so they can be looked up easily.
+        if Options.is_debug and "HIDE_SOURCE" not in flags:
+            source_path = os.path.join(
+                OutputDirectories.getSourceDirectoryPath(), module_name + ".py"
+            )
+
+            putTextFileContents(filename=source_path, contents=code)
+
+        try:
+            trigger_module, _added = buildModule(
+                module_filename=os.path.join(
+                    os.path.dirname(module.getCompileTimeFilename()),
+                    module_name.asPath() + ".py",
+                ),
+                module_package=module_name.getPackageName(),
+                source_code=code,
+                is_top=False,
+                is_main=False,
+                is_shlib=False,
+                is_fake=module_name,
+                hide_syntax_error=False,
+            )
+        except SyntaxError:
+            plugins_logger.sysexit(
+                "SyntaxError in plugin provided source code for '%s'." % module_name
+            )
+
+        if trigger_module.getCompilationMode() == "bytecode":
+            trigger_module.setSourceCode(code)
+
+        return trigger_module
+
+    @classmethod
+    def onModuleDiscovered(cls, module):
+        full_name = module.getFullName()
+
+        def _untangleLoadDesc(descs):
+            if descs and inspect.isgenerator(descs):
+                descs = tuple(descs)
+
+            if descs:
+                if type(descs[0]) not in (tuple, list):
+                    descs = [descs]
+
+                for desc in descs:
+                    if len(desc) == 2:
+                        code, reason = desc
+                        flags = ()
+                    else:
+                        code, reason, flags = desc
+                        if type(flags) is str:
+                            flags = (flags,)
+
+                    yield plugin, code, reason, flags
+
+        preload_descs = []
+        postload_descs = []
+
         for plugin in getActivePlugins():
             plugin.onModuleDiscovered(module)
 
+            preload_descs.extend(
+                _untangleLoadDesc(descs=plugin.createPreModuleLoadCode(module))
+            )
+            postload_descs.extend(
+                _untangleLoadDesc(descs=plugin.createPostModuleLoadCode(module))
+            )
+
+        if preload_descs:
+            total_code = []
+            total_flags = OrderedSet()
+
+            for plugin, pre_code, reason, flags in preload_descs:
+                if pre_code:
+                    plugin.info(
+                        "Injecting pre-module load code for module '%s':" % full_name
+                    )
+                    for line in reason.split("\n"):
+                        plugin.info("    " + line)
+
+                    total_code.append(pre_code)
+                    total_flags.update(flags)
+
+            if total_code:
+                assert full_name not in pre_modules
+
+                pre_modules[full_name] = cls._createTriggerLoadedModule(
+                    module=module,
+                    trigger_name="-preLoad",
+                    code="\n\n".join(total_code),
+                    flags=total_flags,
+                )
+
+        if postload_descs:
+            total_code = []
+            total_flags = OrderedSet()
+
+            for plugin, post_code, reason, flags in postload_descs:
+                if post_code:
+                    plugin.info(
+                        "Injecting post-module load code for module '%s':" % full_name
+                    )
+                    for line in reason.split("\n"):
+                        plugin.info("    " + line)
+
+                    total_code.append(post_code)
+                    total_flags.update(flags)
+
+            if total_code:
+                assert full_name not in post_modules
+
+                post_modules[full_name] = cls._createTriggerLoadedModule(
+                    module=module,
+                    trigger_name="-postLoad",
+                    code="\n\n".join(total_code),
+                    flags=total_flags,
+                )
+
     @staticmethod
     def onModuleSourceCode(module_name, source_code):
         assert type(module_name) is ModuleName
         assert type(source_code) is str
 
         for plugin in getActivePlugins():
-            source_code = plugin.onModuleSourceCode(module_name, source_code)
+            new_source_code = plugin.onModuleSourceCode(module_name, source_code)
+            if new_source_code is not None:
+                source_code = new_source_code
+
             assert type(source_code) is str
 
         return source_code
@@ -352,14 +624,26 @@ class Plugins(object):
 
     @staticmethod
     def onModuleEncounter(module_filename, module_name, module_kind):
-        result = False
+        result = None
 
         for plugin in getActivePlugins():
             must_recurse = plugin.onModuleEncounter(
                 module_filename, module_name, module_kind
             )
 
-            result = result or must_recurse
+            if must_recurse is None:
+                continue
+
+            if type(must_recurse) is not tuple and must_recurse not in (True, False):
+                plugin.sysexit(
+                    "Error, onModuleEncounter code failed to return a None or tuple(bool, reason) result."
+                )
+
+            if result is not None:
+                # false alarm, pylint: disable=unsubscriptable-object
+                assert result[0] == must_recurse[0]
+
+            result = must_recurse
 
         return result
 
@@ -472,8 +756,11 @@ class Plugins(object):
                 # We order per plugin, but from the plugins, lets just take a dict
                 # and achieve determism by ordering the files by name.
                 for key, value in sorted(value.items()):
+                    if not key.startswith("nuitka_"):
+                        key = "plugin." + plugin.plugin_name + "." + key
+
                     assert key not in result, key
-                    result["plugin." + plugin.plugin_name + "." + key] = value
+                    result[key] = value
 
         return result
 
@@ -496,6 +783,43 @@ class Plugins(object):
 
         return cls.extra_link_libraries
 
+    @classmethod
+    def onDataComposerResult(cls, blob_filename):
+        for plugin in getActivePlugins():
+            plugin.onDataComposerResult(blob_filename)
+
+    @classmethod
+    def deriveModuleConstantsBlobName(cls, data_filename):
+        result = deriveModuleConstantsBlobName(data_filename)
+
+        return cls.encodeDataComposerName(result)
+
+    @classmethod
+    def encodeDataComposerName(cls, name):
+        if str is not bytes:
+            # Encoding needs to match generated source code output.
+            name = name.encode("latin1")
+
+        for plugin in getActivePlugins():
+            r = plugin.encodeDataComposerName(name)
+
+            if r is not None:
+                name = r
+                break
+
+        return name
+
+    @classmethod
+    def onFunctionAssignmentParsed(cls, function_body, assign_node):
+        module_name = function_body.getParentModule().getFullName()
+
+        for plugin in getActivePlugins():
+            plugin.onFunctionAssignmentParsed(
+                module_name=module_name,
+                function_body=function_body,
+                assign_node=assign_node,
+            )
+
 
 def listPlugins():
     """Print available standard plugins."""
@@ -521,7 +845,11 @@ def listPlugins():
 def isObjectAUserPluginBaseClass(obj):
     """Verify that a user plugin inherits from UserPluginBase."""
     try:
-        return obj is not NuitkaPluginBase and issubclass(obj, NuitkaPluginBase)
+        return (
+            obj is not NuitkaPluginBase
+            and issubclass(obj, NuitkaPluginBase)
+            and not inspect.isabstract(obj)
+        )
     except TypeError:
         return False
 
@@ -626,6 +954,9 @@ def activatePlugins():
         plugin_name2plugin_classes.items()
     ):
         if plugin_name in Options.getPluginsEnabled():
+            if plugin_class.isAlwaysEnabled():
+                plugin_class.warning("Plugin is defined as always enabled.")
+
             if plugin_class.isRelevant():
                 _addActivePlugin(plugin_class, args=True)
             else:
@@ -657,19 +988,48 @@ def lateActivatePlugin(plugin_name, option_values):
     _addActivePlugin(getPluginClass(plugin_name), args=True, force=True)
 
 
-def _addPluginCommandLineOptions(parser, plugin_class):
-    if plugin_class.plugin_name not in plugin_options:
-        option_group = OptionGroup(parser, "Plugin %s" % plugin_class.plugin_name)
-        plugin_class.addPluginCommandLineOptions(option_group)
+def _addPluginCommandLineOptions(parser, plugin_class, data_files_tags):
+    plugin_name = plugin_class.plugin_name
+
+    if plugin_name not in plugin_options:
+        option_group = OptionGroup(parser, "Plugin %s" % plugin_name)
+        try:
+            plugin_class.addPluginCommandLineOptions(option_group)
+        except OptionConflictError as e:
+            for other_plugin_name, other_plugin_option_list in plugin_options.items():
+                for other_plugin_option in other_plugin_option_list:
+                    # no public interface for that, pylint: disable=protected-access
+                    if (
+                        e.option_id in other_plugin_option._long_opts
+                        or other_plugin_option._short_opts
+                    ):
+                        plugins_logger.sysexit(
+                            "Plugin '%s' failed to add options due to conflict with '%s' from plugin '%s."
+                            % (plugin_name, e.option_id, other_plugin_name)
+                        )
 
         if option_group.option_list:
             parser.add_option_group(option_group)
-            plugin_options[plugin_class.plugin_name] = option_group.option_list
+            plugin_options[plugin_name] = option_group.option_list
         else:
-            plugin_options[plugin_class.plugin_name] = ()
+            plugin_options[plugin_name] = ()
+
+        plugin_data_files_tags = plugin_class.getTagDataFileTagOptions()
+
+        if plugin_data_files_tags:
+            for tag_name, tag_desc in plugin_data_files_tags:
+                if tag_name in (tag for tag, _desc in data_files_tags):
+                    plugins_logger.sysexit(
+                        "Plugin '%s' provides data files tag handling '%s' already provided."
+                        % (plugin_name, tag_name)
+                    )
+
+                data_files_tags.append((tag_name, tag_desc))
+
+                plugin_datatag2pluginclasses[tag_name] = plugin_class
 
 
-def addPluginCommandLineOptions(parser, plugin_names):
+def addPluginCommandLineOptions(parser, plugin_names, data_files_tags):
     """Add option group for the plugin to the parser.
 
     Notes:
@@ -683,12 +1043,16 @@ def addPluginCommandLineOptions(parser, plugin_names):
     """
     for plugin_name in plugin_names:
         plugin_class = getPluginClass(plugin_name)
-        _addPluginCommandLineOptions(parser, plugin_class)
+        _addPluginCommandLineOptions(
+            parser=parser, plugin_class=plugin_class, data_files_tags=data_files_tags
+        )
 
 
-def addUserPluginCommandLineOptions(parser, filename):
+def addUserPluginCommandLineOptions(parser, filename, data_files_tags):
     plugin_class = loadUserPlugin(filename)
-    _addPluginCommandLineOptions(parser, plugin_class)
+    _addPluginCommandLineOptions(
+        parser=parser, plugin_class=plugin_class, data_files_tags=data_files_tags
+    )
 
     user_plugins.add(plugin_class)
 
