@@ -24,6 +24,7 @@ stuff. It will also frequently add sorting for determism.
 
 from __future__ import print_function
 
+import errno
 import glob
 import os
 import shutil
@@ -33,14 +34,15 @@ import time
 from contextlib import contextmanager
 
 from nuitka.__past__ import (  # pylint: disable=I0021,redefined-builtin
+    WindowsError,
     basestring,
 )
 from nuitka.PythonVersions import python_version
-from nuitka.Tracing import my_print
+from nuitka.Tracing import my_print, options_logger
 
 from .Importing import importFromInlineCopy
 from .ThreadedExecutor import RLock, getThreadIdent
-from .Utils import getOS
+from .Utils import getOS, isWin32Windows
 
 # Locking seems to be only required for Windows mostly, but we can keep
 # it for all.
@@ -137,6 +139,15 @@ def relpath(path, start="."):
         raise
 
 
+def isRelativePath(path):
+    if os.path.isabs(path):
+        return False
+    if path.startswith((".." + os.path.sep, "../")):
+        return False
+
+    return True
+
+
 def makePath(path):
     """Create a directory if it doesn't exist.
 
@@ -154,6 +165,11 @@ def makePath(path):
             os.makedirs(path)
 
 
+def isPathExecutable(path):
+    """Is the given path executable."""
+    return os.path.isfile(path) and os.access(path, os.X_OK)
+
+
 def _getRealPathWindows(path):
     # Slow, because we are using an external process, we it's only for standalone and Python2,
     # which is slow already.
@@ -218,14 +234,22 @@ def listDir(path):
     )
 
 
-def getFileList(path, ignore_dirs=(), ignore_suffixes=()):
+def getFileList(
+    path,
+    ignore_dirs=(),
+    ignore_filenames=(),
+    ignore_suffixes=(),
+    only_suffixes=(),
+    normalize=True,
+):
     """Get all files below a given path.
 
     Args:
         path: directory to create a recurseive listing from
         ignore_dirs: Don't descend into these directory, ignore them
+        ignore_filenames: Ignore files named exactly like this
         ignore_suffixes: Don't return files with these suffixes
-
+        only_suffixes: If not empty, limit returned files to these suffixes
 
     Returns:
         Sorted list of all filenames below that directory,
@@ -235,28 +259,50 @@ def getFileList(path, ignore_dirs=(), ignore_suffixes=()):
         This function descends into directories, but does
         not follow symlinks.
     """
+    # We work with a lot of details here, pylint: disable=too-many-locals
+
     result = []
 
+    # Normalize ignoredirs for better matching.
+    ignore_dirs = [os.path.normcase(ignore_dir) for ignore_dir in ignore_dirs]
+    ignore_filenames = [
+        os.path.normcase(ignore_filename) for ignore_filename in ignore_filenames
+    ]
+
     for root, dirnames, filenames in os.walk(path):
         dirnames.sort()
         filenames.sort()
 
         # Normalize dirnames for better matching.
-        dirnames[:] = [os.path.normcase(dirname) for dirname in dirnames]
-        for dirname in ignore_dirs:
-            if dirname in dirnames:
-                dirnames.remove(dirname)
+        dirnames_normalized = [os.path.normcase(dirname) for dirname in dirnames]
+        for ignore_dir in ignore_dirs:
+            if ignore_dir in dirnames_normalized:
+                dirnames.remove(ignore_dir)
+
+        # Normalize filenames for better matching.
+        filenames_normalized = [os.path.normcase(filename) for filename in filenames]
+        for ignore_filename in ignore_filenames:
+            if ignore_filename in filenames_normalized:
+                filenames.remove(ignore_filename)
 
         for filename in filenames:
             if os.path.normcase(filename).endswith(ignore_suffixes):
                 continue
 
-            result.append(os.path.normpath(os.path.join(root, filename)))
+            if only_suffixes and not os.path.normcase(filename).endswith(only_suffixes):
+                continue
+
+            fullname = os.path.join(root, filename)
+
+            if normalize:
+                fullname = os.path.normpath(fullname)
+
+            result.append(fullname)
 
     return result
 
 
-def getSubDirectories(path):
+def getSubDirectories(path, ignore_dirs=()):
     """Get all directories below a given path.
 
     Args:
@@ -273,7 +319,15 @@ def getSubDirectories(path):
 
     result = []
 
+    ignore_dirs = [os.path.normcase(ignore_dir) for ignore_dir in ignore_dirs]
+
     for root, dirnames, _filenames in os.walk(path):
+        # Normalize dirnames for better matching.
+        dirnames_normalized = [os.path.normcase(dirname) for dirname in dirnames]
+        for ignore_dir in ignore_dirs:
+            if ignore_dir in dirnames_normalized:
+                dirnames.remove(ignore_dir)
+
         dirnames.sort()
 
         for dirname in dirnames:
@@ -305,14 +359,22 @@ def deleteFile(path, must_exist):
 
 
 def splitPath(path):
-    """ Split path, skipping empty elements. """
+    """Split path, skipping empty elements."""
     return tuple(element for element in os.path.split(path) if element)
 
 
-def hasFilenameExtension(path, extensions):
-    """ Has a filename one of the given extensions. """
+def getFilenameExtension(path):
+    """Get the filename extension.
 
-    extension = os.path.splitext(os.path.normcase(path))[1]
+    Note: For checks on extension, use hasFilenameExtension instead.
+    """
+    return os.path.splitext(os.path.normcase(path))[1]
+
+
+def hasFilenameExtension(path, extensions):
+    """Has a filename one of the given extensions."""
+
+    extension = getFilenameExtension(path)
 
     return extension in extensions
 
@@ -363,15 +425,30 @@ def getFileContentByLine(filename, mode="r", encoding=None):
 
 
 def getFileContents(filename, mode="r", encoding=None):
-    with withFileLock("reading file %s" % filename):
-        if encoding is not None:
-            import codecs
+    """Get the contents of a file.
 
-            with codecs.open(filename, mode, encoding=encoding) as f:
-                return f.read()
-        else:
-            with open(filename, mode) as f:
-                return f.read()
+    Args:
+        filename: str with the file to be read
+        mode: "r" for str, "rb" for bytes result
+        encoding: optional encoding to used when reading the file, e.g. "utf8"
+
+    Returns:
+        str or bytes - depending on mode.
+
+    """
+
+    with withFileLock("reading file %s" % filename):
+        with openTextFile(filename, mode, encoding=encoding) as f:
+            return f.read()
+
+
+def openTextFile(filename, mode, encoding=None):
+    if encoding is not None:
+        import codecs
+
+        return codecs.open(filename, mode, encoding=encoding)
+    else:
+        return open(filename, mode)
 
 
 def putTextFileContents(filename, contents, encoding=None):
@@ -394,14 +471,8 @@ def putTextFileContents(filename, contents, encoding=None):
                 print(line, file=output_file)
 
     with withFileLock("writing file %s" % filename):
-        if encoding is not None:
-            import codecs
-
-            with codecs.open(filename, "w", encoding=encoding) as output_file:
-                _writeContents(output_file)
-        else:
-            with open(filename, "w") as output_file:
-                _writeContents(output_file)
+        with openTextFile(filename, "w", encoding=encoding) as output_file:
+            _writeContents(output_file)
 
 
 @contextmanager
@@ -467,16 +538,51 @@ def copyTree(source_path, dest_path):
     # False alarm on travis, pylint: disable=I0021,import-error,no-name-in-module
     from distutils.dir_util import copy_tree
 
-    copy_tree(source_path, dest_path)
+    return copy_tree(source_path, dest_path)
+
+
+def copyFileWithPermissions(source_path, dest_path):
+    """Improved version of shutil.copy2.
+
+    File systems might not allow to transfer extended attributes, which we then ignore
+    and only copy permissions.
+    """
+
+    try:
+        shutil.copy2(source_path, dest_path)
+    except PermissionError as e:
+        if e.errno != errno.EACCES:
+            raise
+
+        source_mode = os.stat(source_path).st_mode
+        shutil.copy(source_path, dest_path)
+        os.chmod(dest_path, source_mode)
+
+
+def getWindowsDrive(path):
+    """Windows drive for a given path."""
+
+    drive, _ = os.path.splitdrive(os.path.abspath(path))
+    return os.path.normcase(drive)
 
 
 def isPathBelow(path, filename):
+    """Is a path inside of a given directory path."""
     path = os.path.abspath(path)
     filename = os.path.abspath(filename)
 
+    if isWin32Windows():
+        if getWindowsDrive(path) != getWindowsDrive(filename):
+            return False
+
     return os.path.relpath(filename, path).split(os.path.sep)[0] != ".."
 
 
+def isPathBelowOrSameAs(path, filename):
+    """Is a path inside of a given directory path or the same path as that directory."""
+    return isPathBelow(path, filename) or areSamePaths(path, filename)
+
+
 def getWindowsShortPathName(filename):
     """Gets the short path name of a given long path.
 
@@ -500,7 +606,9 @@ def getWindowsShortPathName(filename):
     output_buf_size = 0
     while True:
         output_buf = ctypes.create_unicode_buffer(output_buf_size)
-        needed = GetShortPathNameW(filename, output_buf, output_buf_size)
+        needed = GetShortPathNameW(
+            os.path.abspath(filename), output_buf, output_buf_size
+        )
 
         if needed == 0:
             # Windows only code, pylint: disable=I0021,undefined-variable
@@ -585,6 +693,44 @@ def replaceFileAtomic(source_path, dest_path):
 
 
 def resolveShellPatternToFilenames(pattern):
-    result = glob.glob(pattern)
+    """Resolve shell pattern to filenames.
+
+    Args:
+        pattern - str
+
+    Returns:
+        list - filenames that matched.
+    """
+
+    if "**" in pattern:
+        if python_version >= 0x350:
+            result = glob.glob(pattern, recursive=True)
+        else:
+            glob2 = importFromInlineCopy("glob2", must_exist=True)
+
+            if glob2 is None:
+                options_logger.sysexit(
+                    "Using pattern with ** is not supported before Python 3.5 unless glob2 is installed."
+                )
+
+            result = glob2.glob(pattern)
+    else:
+        result = glob.glob(pattern)
+
+    result = [os.path.normpath(filename) for filename in result]
     result.sort()
     return result
+
+
+@contextmanager
+def withDirectoryChange(path, allow_none=False):
+    """Change current directory temporarily in a context."""
+
+    if path is not None or not allow_none:
+        old_cwd = os.getcwd()
+        os.chdir(path)
+
+    yield
+
+    if path is not None or not allow_none:
+        os.chdir(old_cwd)
